unit MiniAwk; {
  15feb96 kci -- develop TAwkParser, a simple awk-like line parser

notes 24feb96 kci
This awk object is but a limited subset of a real awk system, but does provide the basic
"parse-input-into-fields" that is a key part of awk.

Example using MiniAwk:
      var arg: TAwkParser;
      ...
      arg := TAwkParser.Create;
      arg.FS := ';'; // otherwise FS defaults to '', which matches any number of spaces
      arg.Parse('any text; to be parsed; on the FS delimiter');
   // arg.CreateParse(';', 'any text; ...');  // would be equivalent to get to this point
      write('the entire input string was "' + arg[0] + '"');
      for i := 1 to arg.NF do
        write('field #' + i + ' is "' + arg[i] + '"');
      arg.Free;

Possible directions:
       * support stream i/o
       * re-parse if any input field is assigned-to
       * support regular expressions
       * support other awkish behavior
}

interface

type
  TAwkParser = class { very simple awk-like line parser}
    private
      FieldSep: string;
      FLine: string;
      NumFields: byte;
      Index, Count: array[0..255] of integer;
      function GetArg(n: byte): string;
    public
      constructor CreateParse(const fs, line: string);
      property FS: string read FieldSep write FieldSep;
      property NF: byte read NumFields;
      property Arg[n: byte]: string read GetArg; default;
      procedure Parse(const Line: string);
    end;

implementation

constructor TAwkParser.CreateParse(const fs, line: string);
  begin
  inherited Create;
  FieldSep := fs;
  Parse(line);
  end;

function TAwkParser.GetArg(n: byte): string;
  begin
  result := '';
  if n > NumFields then
    exit;
  if n = 0 then
    result := FLine
  else if n > 0 then begin
    dec(n); {0-based arrays!}
    if Index[n] > 0 then {count[n] is assumed to contain a valid count, at least 1!}
      result := copy(FLine, index[n], count[n]);
    end;
  end;

procedure TAwkParser.Parse(const line: string);
  var
    n, i, len: integer;
    multFS, inArg: boolean;
    fsep: char;
    s: string;
  begin
  multFS := FieldSep = '';
  if multFS then
    fsep := ' ' { special case: multiple blanks treated as one!}
  else if (FieldSep[1] = '^') and (length(FieldSep) > 1) then
    if FieldSep[2] in ['A'..'[', 'a'..'z'] then
      fsep := chr(ord(FieldSep[2]) - 64) { standard control character}
    else if FieldSep[2] = '^' then 
      fsep := '^' { literal ^}
    else
      fsep := '^' { don't know what else to do... could raise an exception i suppose, but...}
  else
    fsep := FieldSep[1];
  FLine := line;
  len := length(line);
  i := 1;
  n := 0;
  inArg := false;
  count[0] := 0;
  index[0] := 0; {this is used as indicator of null arg}
  while i <= len do begin { parse line}
    if line[i]=fsep then
      if inArg then begin {transition out of arg}
        inArg := false;
        inc(n);
        index[n] := 0; {this is used as indicator of null arg}
        end
      else begin {fsep & not inArg, i.e., last char was fsep too}
        if multFS then
          {special case: treat multiple separators as one}
        else begin {arg[n] is null}
          index[n] := i;
          count[n] := 0;
          inc(n);
          index[n] := 0; {this is used as indicator of null arg}
          end
        end
    else if not inArg then begin {transition into arg}
      inArg := true;
      index[n] := i;
      count[n] := 1;
      end
    else {step along inArg}
      inc(count[n]);
    inc(i); {next char in line}
    end;
  if multFS and not inArg then
    dec(n); { special case: don't count trailing spaces!} 
  if index[0] > 0 then
    NumFields := n + 1
  else
    NumFields := 0
  end;

end.
 